/* 

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include "rtl.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "real.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-flags.h"
#include "output.h"
#include "insn-attr.h"
#include "flags.h"
#include "recog.h"
#include "reload.h"
#include "tree.h"
#include "expr.h"
#include "obstack.h"
#include "except.h"
#include "function.h"
#include "toplev.h"
#include "fcpu.h"

static rtx comparisonOperand0 = 0, 
           comparisonOperand1 = 0;

int
fcpu_comparison_operator (rtx operand, enum machine_mode mode)
{
   if (comparison_operator (operand, mode) == 0)
     return 0;

   switch (GET_CODE (operand))
   {
   case NE:
   case LT:
   case LTU:
   case LE:
   case LEU:
      return 1;

   default:
      return 0;
   }
}
         
int
fcpu_label_operand (rtx operand, enum machine_mode mode)
{
   return (GET_CODE (operand) == CODE_LABEL ||
           GET_CODE (operand) == LABEL_REF);
}

int
fcpu_jump_operand (rtx operand, enum machine_mode mode)
{
   return fcpu_label_operand (operand, mode) ||
          register_operand (operand, mode);
}

rtx
fcpu_jump_emit (rtx codeLabel)
{ 
   return force_reg (Pmode, gen_rtx_LABEL_REF (Pmode, codeLabel));
}

rtx
fcpu_zero_extend_emit (rtx result, rtx source)
{
    enum machine_mode extendMode = GET_MODE (source),
                      resultMode = GET_MODE (result);
    rtx destination              = gen_rtx_SUBREG (extendMode, result, 0);

    emit_move_insn (destination, source);

    emit_insn (gen_rtx_SET (VOIDmode,
                            result,
                            gen_rtx_ZERO_EXTEND (resultMode, destination)));

    return result;
}

rtx
fcpu_move_emit (rtx result, rtx source)
{
    enum machine_mode moveMode = GET_MODE (result);

    if (GET_CODE (result) != REG ||
        (moveMode != SImode &&
         moveMode != DImode &&
         moveMode != SFmode &&
         moveMode != DFmode) ||
        CONSTANT_P (source) == 0 ||
        GET_CODE (source) == LABEL_REF ||
        (GET_CODE (source) == CONST_INT &&
         (CONST_OK_FOR_LETTER_P (INTVAL (source), 'I') != 0 ||
          CONST_OK_FOR_LETTER_P (INTVAL (source), 'P') != 0 ||
          (GET_MODE_SIZE (moveMode) <= 2 &&
           CONST_OK_FOR_LETTER_P (INTVAL (source), 'J')))))
      return NULL_RTX;
    
    emit_insn (gen_loadcons (result, result, source, const0_rtx));
    emit_insn (gen_loadcons (result, result, source, GEN_INT (1)));
    
    if (moveMode == DImode ||
        moveMode == DFmode)
    {
        emit_insn (gen_loadcons (result, result, source, GEN_INT (2)));
        emit_insn (gen_loadcons (result, result, source, GEN_INT (3)));
    }

    return result;
} 

void
fcpu_compare_operands (const rtx * ops)
{
    comparisonOperand0 = ops [0];
    comparisonOperand1 = ops [1];
}

rtx
fcpu_compare_set_emit (enum rtx_code compareCode, rtx result)
{
    enum machine_mode compareMode = GET_MODE (comparisonOperand0);
    rtx setResult;

    if (compareMode != EQ &&
        compareMode != NE &&
        (compareMode == SFmode ||
         compareMode == DFmode))
    {
        enum rtx_code actualCode;
        enum machine_mode setMode;
        rtx compareResult;

        switch (compareCode)
        {
        case GE:
        case GT:
           actualCode = swap_condition (compareCode);

           break;

        default:
           actualCode = compareCode;
           
           break;        
        }

        compareResult = gen_reg_rtx (compareMode);

        if (actualCode == compareCode)
          emit_insn (gen_rtx_SET (VOIDmode,
                                  compareResult,
                                  gen_rtx_MINUS (compareMode,
                                                 comparisonOperand0,
                                                 comparisonOperand1)));
        else
          emit_insn (gen_rtx_SET (VOIDmode,
                                  compareResult,
                                  gen_rtx_MINUS (compareMode,
                                                 comparisonOperand1,
                                                 comparisonOperand0)));

        setMode   = (compareMode == DFmode) ? DImode : SImode;
        setResult = gen_reg_rtx (setMode);

        emit_insn (gen_float_expand (setResult, compareResult));
        
        if (actualCode == LE)
          emit_insn (gen_rtx_SET (VOIDmode,
                                  setResult,
                                  gen_rtx_MINUS (setMode, 
                                                 setResult, 
                                                 GEN_INT (1))));
        emit_insn (gen_rtx_SET (VOIDmode,
                                setResult,
                                gen_rtx_LSHIFTRT (setMode,
                                                  setResult,
                                                  GEN_INT (GET_MODE_SIZE (setMode) * 8 - 1))));

    }
    else
    switch (compareCode)
    {
    case EQ:
         {
             setResult = gen_reg_rtx (compareMode);

             emit_insn (gen_rtx_SET (VOIDmode,
                                     setResult,
                                     gen_rtx_fmt_ee (NE,
                                                     compareMode,
                                                     comparisonOperand0,
                                                     comparisonOperand1)));
             
             setResult = gen_rtx_fmt_ee (EQ, compareMode, setResult, const0_rtx);
         }

         break;

    case GT:
    case GE:
    case GTU:
    case GEU:
         compareCode = swap_condition (compareCode);
         
         setResult = gen_rtx_fmt_ee (compareCode,
                                     compareMode,
                                     comparisonOperand1,
                                     comparisonOperand0);
    
         break;

    default:
                                
         setResult = gen_rtx_fmt_ee (compareCode, 
                                     compareMode,
                                     comparisonOperand0,
                                     comparisonOperand1);

         break;
    }

    if (GET_MODE_SIZE (GET_MODE (setResult)) != GET_MODE_SIZE (GET_MODE (result)))
    {
         rtx convertResult = setResult;

         if (GET_CODE (setResult) != REG)
         {
            convertResult = gen_reg_rtx (GET_MODE (setResult));

            emit_insn (gen_rtx_SET (GET_MODE (setResult), convertResult, setResult));
         }

         if (GET_MODE_SIZE (GET_MODE (convertResult)) < GET_MODE_SIZE (GET_MODE (result)))
           setResult = gen_rtx_SIGN_EXTEND (GET_MODE (result), convertResult);
         else
           setResult = gen_rtx_TRUNCATE (GET_MODE (result), convertResult);
    }

    return setResult;
}

rtx
fcpu_compare_move_emit (enum rtx_code compareCode)
{
    return fcpu_compare_jump_emit (compareCode);
}

rtx
fcpu_compare_jump_emit (enum rtx_code compareCode)
{
    enum rtx_code actualCode;
    enum machine_mode compareMode = GET_MODE (comparisonOperand0);
    rtx compareResult             = gen_reg_rtx (compareMode);
  
    if (compareCode != EQ &&
        compareCode != NE &&
        (compareMode == SFmode ||
         compareMode == DFmode))
    {
        rtx extendResult;
        
        switch (compareCode)
        {
        case GT:
           actualCode = GE;
        
           break;
           
        case LE:
           actualCode = LT;

           break;
           
        default:
           actualCode = compareCode;

           break;
        }

        emit_insn (gen_rtx_SET (VOIDmode,
                                compareResult,
                                gen_rtx_MINUS (compareMode,
                                               comparisonOperand0,
                                               comparisonOperand1)));
    
        extendResult = gen_reg_rtx (DImode);

        emit_insn (gen_float_expand (extendResult, compareResult));

        if (compareCode == GT ||
            compareCode == LE)
          emit_insn (gen_rtx_SET (VOIDmode,
                                  extendResult,
                                  gen_rtx_MINUS (DImode,
                                                 extendResult,
                                                 GEN_INT (1))));
       
        return gen_rtx_fmt_ee ((actualCode == LT) ? NE : EQ,
                               DImode,
                               gen_rtx_LSHIFTRT (DImode, extendResult, GEN_INT (63)),
                               const0_rtx);
    }
                               
    switch (compareCode)
    {
    case EQ:
    case GT:
    case GE:
    case GTU:
    case GEU:
        actualCode = reverse_condition (compareCode);
            
        break;

    default:
        actualCode = compareCode;

        break;
    }
  
    emit_insn (gen_rtx_SET (VOIDmode,
                            compareResult,
                            gen_rtx_fmt_ee (actualCode,
                                            compareMode,
                                            comparisonOperand0,
                                            comparisonOperand1)));
  
    if (compareMode != DImode)
    {
        rtx extendResult = gen_reg_rtx (DImode);

        emit_insn (gen_rtx_SET (VOIDmode,
                                extendResult,
                                gen_rtx_SIGN_EXTEND (DImode, compareResult)));

        compareResult = extendResult;
    }

    return gen_rtx_fmt_ee ((compareCode == actualCode) ? EQ : NE, 
                           DImode, 
                           compareResult, 
                           CONST0_RTX (compareMode));
}

const char *
fcpu_compare_move_output (enum rtx_code compareCode, rtx condition, rtx result, rtx source)
{
    const char * size  = "",
               * flags = "";
    char template [128];
    rtx operands [3] = { condition, source, result };

    switch (GET_MODE (result))
    {
    case SFmode:
    case SImode:
         size = "32";

         break;

    case HImode:
         size = "16";

         break;

    case QImode:
         size = "8";

         break;
    }
    
    if (compareCode == NE)
      flags = ".n";

    sprintf (template, "move%s%s\t%%0, %%1, %%2", size, flags);
    
    output_asm_insn (template, operands);

    return "";
}

const char * 
fcpu_compare_set_output (enum rtx_code compareCode, rtx result, rtx operand0, rtx operand1)
{
    const char * size   = "",
               * flags  = "",
               * suffix = "";
    char template [128];
    rtx operands [3] = { operand0, operand1, result };

    switch (GET_MODE (operand0))
    {
    case SFmode:
    case SImode:
         size = "32";
         
         break;
    
    case HImode:
         size = "16";
         
         break;
    
    case QImode:
         size = "8";
         
         break;
    }

    if (compareCode == LT ||
        compareCode == LE)
      flags = ".s";
   
    if (operand0 == const0_rtx)
      operands [0] = gen_rtx_REG (GET_MODE (operand1), 0);
    else
    if (operand1 == const0_rtx)
      operands [1] = gen_rtx_REG (GET_MODE (operand0), 0);
    else
      suffix = "i";
    
    switch (compareCode)
    {
    case NE:
      sprintf (template, "xor%s%s\t%%1, %%0, %%2", suffix, size);
      
      break;
    
    case LT:
    case LTU:
      sprintf (template, "cmpl%s%s%s\t%%1, %%0, %%2", suffix, size, flags);
      
      break;      
    
    case LE:
    case LEU:
      sprintf (template, "cmple%s%s%s\t%%1, %%0, %%2", suffix, size, flags);
      
      break;

    default:
      abort ();
    }

    output_asm_insn (template, operands);
    
    return "";
}

const char *
fcpu_float_expand_output (rtx result, rtx source)
{
    rtx operands [2] = { result, source };

    if (GET_MODE (source) == SFmode)
    {
      if (GET_MODE (result) == DImode)
        output_asm_insn ("move32.s\tr0, %1, %0", operands);
      else
      if (REGNO (result) != REGNO (source))
        output_asm_insn ("move32\tr0, %1, %0", operands);
    }
    else
    if (REGNO (result) != REGNO (source))
      output_asm_insn ("move\tr0, %1, %0", operands);

    return "";
}

void 
fcpu_function_prologue (FILE * file, int allocateSize)
{
   extern char   call_used_regs [];
   extern char * reg_names [];

   int currentRegister, spillSize;
  
   spillSize = fcpu_spill_size ();
   if (spillSize % 16)
     allocateSize += 8;
   if (allocateSize % 16)
     allocateSize += 8;
   
   if (spillSize <= 0)
   {
     if (allocateSize > 0)
       fprintf (file, "\tsubi\t%d, r63, r63\n", allocateSize);

     return;
   }

   fprintf (file, "\tsubi\t%d, r63, r63\n", allocateSize + 8);

   for (currentRegister = 0;
        spillSize > 8;
        spillSize -= 8, 
          ++ currentRegister)
   {
      while (regs_ever_live [currentRegister] == 0 ||
             (call_used_regs [currentRegister] != 0 &&
              currentRegister != RETURN_ADDRESS_REGNUM))
         ++ currentRegister;

      fprintf (file, "\tstorei\t-8, r63, %s\n", reg_names [currentRegister]);
   }

   if (frame_pointer_needed != 0)
     currentRegister = FRAME_POINTER_REGNUM;
   else
   if (regs_ever_live [RETURN_ADDRESS_REGNUM] != 0)
     currentRegister = RETURN_ADDRESS_REGNUM;
   else
   while (regs_ever_live [currentRegister] == 0||
	      call_used_regs [currentRegister] != 0)
      ++ currentRegister;
   
   fprintf (file, "\tstore\tr63, %s\n", reg_names [currentRegister]);

   if (frame_pointer_needed != 0)
     fprintf (file, "\tmove\tr0, r63, r61\n");
}

void 
fcpu_function_epilogue (FILE * file, int allocateSize)
{
   extern char   call_used_regs [];
   extern char * reg_names [];

   int currentRegister, spillSize;

   if (frame_pointer_needed != 0)
     fprintf (file, "\tmove\tr0, r61, r63\n");

   spillSize = fcpu_spill_size ();
   if (spillSize % 16)
     allocateSize += 8;
   if (allocateSize % 16)
     allocateSize += 8;

   if (spillSize <= 0)
   {
     if (allocateSize > 0)
       fprintf (file, "\taddi\t%d, r63, r63\n", allocateSize);
     
     return;
   }

   if (spillSize > 8 &&
       frame_pointer_needed != 0)
   {
      fprintf (file, "\tload\t8, r63, r61\n");

      spillSize -= 8;
   }
   
   for (currentRegister = FIRST_PSEUDO_REGISTER - 1;
        spillSize > 8;
        spillSize -= 8,
           -- currentRegister)
   {
      while (regs_ever_live [currentRegister] == 0 ||
             (call_used_regs [currentRegister] != 0 &&
              currentRegister != RETURN_ADDRESS_REGNUM))
         -- currentRegister;
      
      fprintf (file, "\tloadi\t8, r63, %s\n", reg_names [currentRegister]);
   }

   if (frame_pointer_needed != 0 &&
       currentRegister >= FIRST_PSEUDO_REGISTER - 1)
     currentRegister = FRAME_POINTER_REGNUM;
   else
   while (regs_ever_live [currentRegister] == 0 ||
          (call_used_regs [currentRegister] != 0 &&
           currentRegister != RETURN_ADDRESS_REGNUM))
      -- currentRegister;      
   
   fprintf (file, 
            "\tloadi\t%d, r63, %s\n",
            allocateSize + 8,
            reg_names [currentRegister]);

   fprintf (file, "\tjmpa\tr0, r60, r0\n");
}

void
fcpu_print_operand (FILE * file, rtx operand, int code)
{
   switch (GET_CODE (operand))
   {
   case CONST_DOUBLE:
      if (GET_MODE (operand) == SFmode &&
          CONST_DOUBLE_HIGH (operand) != 0)
      {
         HOST_WIDE_INT hex;
         REAL_VALUE_TYPE real;

         REAL_VALUE_FROM_CONST_DOUBLE (real, operand);
         REAL_VALUE_TO_TARGET_SINGLE (real, hex);

         fprintf (file, HOST_WIDE_INT_PRINT_HEX, hex);
      }
      else
      if (CONST_DOUBLE_HIGH (operand) != 0)
        fprintf (file, 
                 HOST_WIDE_INT_PRINT_DOUBLE_HEX,
                 CONST_DOUBLE_HIGH (operand), 
                 CONST_DOUBLE_LOW (operand));
      else 
      if (CONST_DOUBLE_LOW (operand) < 0)
        fprintf (file, HOST_WIDE_INT_PRINT_HEX, CONST_DOUBLE_LOW (operand));
      else
        fprintf (file, HOST_WIDE_INT_PRINT_DEC, CONST_DOUBLE_LOW (operand));

      break;

   case REG:
      fprintf (file, "%s", reg_names [REGNO (operand)]);

      break;

   case MEM:
      if (code == 'r' &&
          GET_CODE (XEXP (operand, 0)) == REG)
        fprintf (file, "%s", reg_names [REGNO (XEXP (operand, 0))]);
      else
        output_address (XEXP (operand, 0));

      break;
   
   default:
      output_addr_const (file, operand);

      break;
   }
}
          
void
fcpu_print_operand_address (FILE * file, rtx address)
{
   switch (GET_CODE (address))
   {
   case MEM:
      if (GET_CODE (XEXP (address, 0)) == REG)
        fprintf (file, "%s", reg_names [REGNO (address)]);
      else 
        abort();
     
      break;

   case REG:
      fprintf (file, "%s", reg_names [REGNO (address)]);
      
      break;
      
   default:
      abort ();
   }
}

int
fcpu_spill_size (void)
{
  int spillSize = 0, 
      currentRegister;

  for (currentRegister = 0; 
       currentRegister < FIRST_PSEUDO_REGISTER;
       ++ currentRegister)
     if (call_used_regs [currentRegister] == 0 &&
         regs_ever_live [currentRegister] != 0)
       ++ spillSize;

  if (regs_ever_live [RETURN_ADDRESS_REGNUM] != 0)
    ++ spillSize;

  if (frame_pointer_needed != 0)
    ++ spillSize;

  return spillSize * 8;
}



